using System;
using System.Web;
using System.Web.UI;
using System.Web.UI.WebControls;
using System.ComponentModel;
using System.Collections;
using System.Xml;
using System.Text;

namespace NeoCore.PanelControl {	
	/// <summary>
	/// XMSPanelControl is a control that will allow easier form to NeoCore XMS integration
	/// </summary>
	[DefaultProperty("TabIndex"), ToolboxData("<{0}:XMSPanelControl runat=server></{0}:XMSPanelControl>")]
	public class XMSPanelControl : Panel, XMSControl {

		private static string PATH_SEPARATOR = "/";

		/// <summary>
		/// The machine name of the XMS instance to connect to
		/// </summary>
		private string  machineName;
		[Category("Text")]
		[Description("The Machine Name of the XMS instance to connect to")]
		public string MachineName {
			get {
				return machineName;
			}
			set {
				machineName = value;
			}
		}

		/// <summary>
		/// The port number of the XMS instance to connect to
		/// </summary>
		private int  port;
		[Category("int")]
		[Description("The port number of the XMS instance to connect to")]
		public int Port {
			get {
				return port;
			}
			set {
				port = value;
			}
		}

		/// <summary>
		/// The User Name for the XMS instance to connect to
		/// </summary>
		private string  userName;
		[Category("Text")]
		[Description("The User Name for the XMS instance to connect to")]
		public string UserName {
			get {
				return userName;
			}
			set {
				userName = value;
			}
		}

		/// <summary>
		/// The Password for the XMS instance to connect to
		/// </summary>
		private string  password;
		[Category("Text")]
		[Description("The Password for the XMS instance to connect to")]
		public string Password {
			get {
				return password;
			}
			set {
				password = value;
			}
		}

		/// <summary>
		/// populatePanel populates a panel from an XmlDocument
		/// </summary>
		/// <param name="results">The results being built</param>
		/// <param name="record">The record being populated</param>
		/// <param name="panel">The panel to populate</param>
		public void populatePanel(string results, int record, XMSPanelControl panel) {
			XmlDocument doc = new XmlDocument();
			doc.LoadXml(results);
			populateFields(doc.ChildNodes[0].ChildNodes[record], panel);
		}

		/// <summary>
		/// populatePanel populates a panel from an XmlDocument
		/// </summary>
		/// <param name="doc">The document data source</param>
		/// <param name="recordNumber">The record id to populate (may be changed if more than one exists)</param>
		private void populatePanel(XmlDocument doc, int recordNumber) {
			populateFields(doc.ChildNodes[1].ChildNodes[recordNumber], this);
		}

		/// <summary>
		/// populateFields populates fields within a panel
		/// </summary>
		/// <param name="thisNode">The node to populate fields from</param>
		/// <param name="parentControl">The control equating to the xml node</param>
		private void populateFields(XmlNode thisNode, Control parentControl) {
			if (parentControl is XMSPanelDataGrid) {
				populateControl(parentControl, thisNode.InnerXml);
			}
			else if (thisNode != null && thisNode.HasChildNodes ) 
				foreach (XmlNode node in thisNode.ChildNodes) {
					populateFields(node, (parentControl != null && parentControl.ID.Equals(node.Name) ? 
						parentControl : 
						findControl(parentControl, node.Name)));	
				}
			else if (thisNode != null && parentControl != null)
				populateControl(parentControl, thisNode.Value);
		}

		/// <summary>
		/// findControl locates a control within the tree of controls
		/// </summary>
		/// <param name="parentControl">The control to search</param>
		/// <param name="name">The name of the control being searched for</param>
		/// <returns>The found control</returns>
		private Control findControl(Control parentControl, string name) {
			Control control = null;
			if (parentControl != null) {
				control = parentControl.FindControl(name);
				if (!name.StartsWith("#")) {				
					if (control == null) {
						foreach (Control aControl in parentControl.Controls) {
							control = findControl(aControl, name);
							if (control != null)
								break;
						}
					}
				} 
				else {
					control = parentControl;
				}
			}
			return control;
		}

		/// <summary>
		/// populateControl  The control to populate
		/// </summary>
		/// <param name="thisControl">The control to populate</param>
		/// <param name="data">The data to populate it with</param>
		private void populateControl(Control thisControl, string data) {
			if (thisControl is XMSControl) 
				((XMSControl)thisControl).SetData(data);

		}

		/// <summary>
		/// buildQueryString builds the query string
		/// </summary>
		/// <param name="query">The query to build</param>
		/// <param name="panel">The panel to build the string from</param>
		/// <returns>The query represented by the panels</returns>
		public string buildQueryString(string query, XMSPanelControl panel) {
			return buildQueryString(query, panel, 0);
		}

		/// <summary>
		/// buildQueryString builds the query string for the panel
		/// </summary>
		/// <param name="query">The Query to run</param>
		/// <param name="panel">The panel to load results to</param>
		/// <returns>The results string</returns>
		public String buildQueryString(String query, XMSPanelControl panel, int depth) {
			StringBuilder xpath = new StringBuilder("/ND/");
			xpath.Append(panel.ID);
			xpath.Append("[");
			xpath.Append(panel.ToXPathPredicate("root"));
			xpath.Append("]");

			return xpath.ToString();
		}

		/// <summary>
		/// traverseQueryControl traverses the query, sending back the query data
		/// </summary>
		/// <param name="query">The query to append to</param>
		/// <param name="parentControl"The control to traverse></param>
		/// <param name="depth">the depth of the traverse</param>
		/// <returns>The query string</returns>
		private string traverseQueryControl(string query, Control parentControl, int depth) {
			foreach (Control childControl in parentControl.Controls) {
				if (childControl is XMSPanelControl) {
					query = buildQueryString(query, (XMSPanelControl)childControl, depth + 1);
				} else {
					query = traverseQueryControl(query, childControl, depth);
				}
			}
			return query;
		}

		/// <summary>
		/// traverseClearControl walks through the controls found in this one, looking for XMSControls to clear
		/// </summary>
		/// <param name="parentControl">The control to traverse</param>
		private void traverseClearControl(Control parentControl) {
			foreach (Control childControl in parentControl.Controls) {
				if (childControl is XMSPanelControl) {
					((XMSPanelControl)childControl).Clear((XMSPanelControl)childControl);
				} else {
					traverseClearControl(childControl);
				}
			}
		}

		/// <summary>
		/// Clear data from the control
		/// </summary>
		public void Clear() {
			Clear(this);
		}

		/// <summary>
		/// Clear clears the input controls
		/// </summary>
		/// <param name="panel">The Panel to clear</param>
		public void Clear(XMSPanelControl panel) {
			ControlCollection controls = panel.Controls;
			IEnumerator denum = controls.GetEnumerator();

			while(denum.MoveNext()) {
				Control tmp = (Control)denum.Current;
				if (tmp is XMSControl) 
					((XMSControl)tmp).Clear();
				else 
					traverseClearControl(tmp);
			}
		}
		
		/// <summary>
		/// Generates an XML representation of this object.
		/// </summary>
		/// <returns>The xml string of the data</returns>
		public string ToXML() {
			StringBuilder sb = new StringBuilder();

			sb.Append("<");
			sb.Append(this.ID);
			sb.Append(">\n");
			
			foreach (Control tmp in this.Controls) {
				if ( tmp is XMSControl ) {
					sb.Append(((XMSControl)tmp).ToXML());
				} else
					sb.Append( traverseNonXMSControl(tmp) );
			}
			sb.Append( "</" + this.ID + ">\n");
			return sb.ToString();
		}

		/// <summary>
		/// traverseNonXMSControl traverses the panel control looking for XMSControls to append values to
		/// </summary>
		/// <param name="parentControl">The parent control to traverse</param>
		/// <returns>A string representing the control data</returns>
		private string traverseNonXMSControl(Control parentControl) {
			StringBuilder csb = new StringBuilder();
			foreach (Control childControl in parentControl.Controls) {
				if (childControl is XMSControl) {
					csb.Append(((XMSControl)childControl).ToXML());
				} else {
					csb.Append(traverseNonXMSControl(childControl));
				}
			}
			return csb.ToString();
		}

		/// <summary>
		/// Populate this object with data from the given parameter.
		/// </summary>
		/// <param name="data"></param>
		public void SetData(string data) {
			System.Xml.XmlDocument doc = new System.Xml.XmlDocument();
			if (data != null)
				doc.LoadXml(data);
			setData(doc);
		}

		/// <summary>
		/// Populate this object with the given XmlDocument
		/// </summary>
		/// <param name="doc">XmlDocument containing data with which to populate the panel</param>
		public void setData(XmlDocument doc) {
			populatePanel(doc, 0);
		}

		/// <summary>
		/// Populate this object with data from the given parameter.
		/// </summary>
		/// <param name="data">The data to set the panel to</param>
		/// <param name="recordNumber">The document may contain several records.</param>
		public void setData(string data, int recordNumber) {
			System.Xml.XmlDocument doc = new System.Xml.XmlDocument();
			if (data != null)
				doc.LoadXml(data);
			setData(doc, recordNumber);
		}

		/// <summary>
		/// Populate this object with the given XmlDocument
		/// </summary>
		/// <param name="doc">XmlDocument containing data with which to populate the panel</param>
		/// <param name="recordNumber">The document may contain several records.</param>
		public void setData(XmlDocument doc, int recordNumber) {
			populatePanel(doc, recordNumber);
		}
		
		/// <summary>
		/// See XMSControl interface
		/// </summary>
		/// <param name="prependValue">Value to prepend to the predicate - e.g. parent panel ID</param>
		/// <returns></returns>
		public string ToXPathPredicate(string prependValue) {
			string predicate = "";
			StringBuilder sb = new StringBuilder();
			prependValue = (prependValue.Equals("root") ? "" : ID);
			foreach (Control ctrl in this.Controls) {
				if (ctrl is XMSControl) {
					predicate = ((XMSControl)ctrl).ToXPathPredicate(prependValue);
					if (predicate.Length > 0 && sb.Length > 0) {
						sb.Append(" and ");
					}
					sb.Append(predicate);
				} else
					sb.Append(traverseNonXMSControlForXPath(ctrl, prependValue));
			}
			return sb.ToString();
		}

		/// <summary>
		/// traverseNonXMSControlForXPath traverses the panel, looking for XMSPanelControl objects
		/// </summary>
		/// <param name="parentControl">The control that I am</param>
		/// <param name="prependValue">The value to prepend to a predicate</param>
		/// <returns></returns>
		private string traverseNonXMSControlForXPath(Control parentControl, string prependValue) {
			StringBuilder csb = new StringBuilder();			
			foreach (Control childControl in parentControl.Controls) {
				if (childControl is XMSControl) {
					csb.Append(((XMSControl)childControl).ToXPathPredicate(prependValue.Equals("root") ? 
						childControl.ID : prependValue));
				} else {
					csb.Append(traverseNonXMSControlForXPath(childControl, prependValue));
				}
			}
			return csb.ToString();
		}
	}
}
